/*
 * This file is part of WCX_SFF. A Windows Commander plugin for SFF files.
 * Copyright (C) 2001  Amenna Heratsini
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * email: amennah@hotmail.com
 */

#include <windows.h>
#include "bmp.h"

#pragma pack(push,1)
static struct {
    char bfTag[2];
    DWORD bfSize;
    WORD bfReserved1;
    WORD bfReserved2;
    DWORD bfOffBits;
} TBITMAPFILEHEADER;

static struct {
    DWORD biSize;           
    LONG biWidth;          
    LONG biHeight;         
    WORD biPlanes;          
    WORD biBitCount;        
    DWORD biCompression;
    DWORD biSizeImage;      
    LONG biXPelsPerMeter;
    LONG biYPelsPerMeter;
    DWORD biClrUsed;
    DWORD biClrImportant;
} TBITMAPINFOHEADER;

static struct {
    DWORD bcSize; 
    WORD bcWidth; 
    WORD bcHeight; 
    WORD bcPlanes; 
    WORD bcBitCount; 
} TBITMAPCOREHEADER;
#pragma pack(pop)

static WORD getbuf, putbuf;
static WORD bmpbytes,bmpptr;
static BYTE bmpbuffer[10000];
static BYTE putval;

static void bmp_put_byte(HANDLE handle, BYTE b)
{
    DWORD written;
    
    if (bmpptr==10000) {
        WriteFile(handle,bmpbuffer,10000,&written,NULL);
        bmpptr=0;
    } 
    bmpbuffer[bmpptr++]=b;
}

static void bmp_flush_buffer(HANDLE handle)
{
    DWORD written;
    
    WriteFile(handle,bmpbuffer,bmpptr,&written,NULL);
    bmpptr=0;
}

// msb first
static void bmp_put_bit(HANDLE handle, BYTE bit)
{    
    if (putbuf==0) bmp_flush_bitbuffer(handle);
    putbuf--;
    putval|=( (bit&0x1)<<putbuf );
}    

static void bmp_flush_bitbuffer(HANDLE handle)
{
    bmp_put_byte(handle, putval);
    putval=0;
    putbuf=8;
}

static unsigned char bmp_get_byte(HANDLE handle)
{
    DWORD read;
    
    if (bmpbytes==0) {
        ReadFile(handle,bmpbuffer,10000,&read,NULL);
        bmpbytes=read;
        bmpptr=0;
    } 
    bmpbytes--;
    return(bmpbuffer[bmpptr++]);
}

// msb first
static BYTE bmp_get_bit(HANDLE handle)
{
    static BYTE getval;
    
    if (getbuf==0) {
        getval=bmp_get_byte(handle);
        getbuf=8;
    }
    getbuf--;
    return(getval>>getbuf&0x1);
}    

static BYTE bmp_get_bits(HANDLE handle, WORD n)
{
    BYTE i;
    BYTE val=0;

    for (i=0;i<n;i++) {
        val=val<<1;
        val=val|bmp_get_bit(handle);
    }
    return(val);
}    

void BMPencode(HANDLE handle, BYTE *image, WORD width, WORD height, WORD bpp)
{
    int x,y;
    BYTE skip;
    DWORD written;
    WORD bytesperline;
    BYTE palette[2][3]={0,0,0,255,255,255};

    if (bpp!=1 && bpp!=24) return;
    
    bytesperline=(width*bpp)/8;
    if ((width*bpp)%8) bytesperline++;

    if (bytesperline%4==0)
        skip=0; else
        skip=4-(bytesperline%4);

    TBITMAPFILEHEADER.bfTag[0]='B';
    TBITMAPFILEHEADER.bfTag[1]='M';
    TBITMAPFILEHEADER.bfSize=bytesperline*height+sizeof(TBITMAPFILEHEADER)+sizeof(TBITMAPINFOHEADER);
    TBITMAPFILEHEADER.bfReserved1=0;
    TBITMAPFILEHEADER.bfReserved2=0;
    TBITMAPFILEHEADER.bfOffBits=sizeof(TBITMAPFILEHEADER)+sizeof(TBITMAPINFOHEADER);
    if (bpp<16) TBITMAPFILEHEADER.bfOffBits+=4*(1<<bpp);
    WriteFile(handle,&TBITMAPFILEHEADER,sizeof(TBITMAPFILEHEADER),&written,NULL);

    TBITMAPINFOHEADER.biSize=40;
    TBITMAPINFOHEADER.biWidth=width;
    TBITMAPINFOHEADER.biHeight=height;
    TBITMAPINFOHEADER.biPlanes=1;
    TBITMAPINFOHEADER.biBitCount=bpp;
    TBITMAPINFOHEADER.biCompression=0;
    TBITMAPINFOHEADER.biSizeImage=0;
    TBITMAPINFOHEADER.biXPelsPerMeter=0;
    TBITMAPINFOHEADER.biYPelsPerMeter=0;
    TBITMAPINFOHEADER.biClrUsed=0;
    TBITMAPINFOHEADER.biClrImportant=0;
    if (bpp<16) TBITMAPINFOHEADER.biClrUsed+=1<<bpp;
    WriteFile(handle,&TBITMAPINFOHEADER,sizeof(TBITMAPINFOHEADER),&written,NULL);

    if (bpp==1) {
        BYTE b=0;
        for (x=0;x<TBITMAPINFOHEADER.biClrUsed;x++) {
            WriteFile(handle,&palette[x][2],1,&written,NULL);
            WriteFile(handle,&palette[x][1],1,&written,NULL);
            WriteFile(handle,&palette[x][0],1,&written,NULL);
            WriteFile(handle,&b,1,&written,NULL);
        }
    }

    bmpptr=0;
    putval=0;
    putbuf=8;

    if (TBITMAPINFOHEADER.biCompression==0 && bpp==1)  // No compression, bpp=1
    for (y=height-1;y>=0;y--) {
        for (x=0;x<width;x++) {
            bmp_put_bit(handle,image[y*width*3+x*3+0]!=0);
        }
        bmp_flush_bitbuffer(handle);
        for (x=0;x<skip;x++) bmp_put_byte(handle, 0);
    } else
    if (TBITMAPINFOHEADER.biCompression==0 && bpp==24)  // No compression, bpp=24
    for (y=height-1;y>=0;y--) {
        for (x=0;x<width;x++) {
            bmp_put_byte(handle,image[y*width*3+x*3+2]);
            bmp_put_byte(handle,image[y*width*3+x*3+1]);
            bmp_put_byte(handle,image[y*width*3+x*3+0]);
        }
        for (x=0;x<skip;x++) bmp_put_byte(handle,0);
    }
    bmp_flush_buffer(handle);
}

void BMPdecode(HANDLE handle, BYTE **image)
{
    int x,y,i;
    BYTE *ptr=*image, skip;
    BYTE ch,ch2,index;
    BYTE palette[256][3];
    WORD width, height, bpp, bytesperline;
    BYTE r,g,b;
    DWORD read, size;
    
    if (BMPinfo(handle, &width, &height, &bpp)<0) return;

    SetFilePointer(handle,0,NULL,FILE_BEGIN);
    ReadFile(handle,&TBITMAPFILEHEADER,sizeof(TBITMAPFILEHEADER),&read,NULL);
    ReadFile(handle,&size,4,&read,NULL);
    SetFilePointer(handle,sizeof(TBITMAPFILEHEADER),NULL,FILE_BEGIN);
    if (size==12) {
        ReadFile(handle,&TBITMAPCOREHEADER,sizeof(TBITMAPCOREHEADER),&read,NULL);
        TBITMAPINFOHEADER.biSize=TBITMAPCOREHEADER.bcSize;
        TBITMAPINFOHEADER.biWidth=abs(TBITMAPCOREHEADER.bcWidth);
        TBITMAPINFOHEADER.biHeight=abs(TBITMAPCOREHEADER.bcHeight);
        TBITMAPINFOHEADER.biPlanes=TBITMAPCOREHEADER.bcPlanes;
        TBITMAPINFOHEADER.biBitCount=TBITMAPCOREHEADER.bcBitCount;
        TBITMAPINFOHEADER.biCompression=0;
        TBITMAPINFOHEADER.biClrUsed=0;
    } else ReadFile(handle,&TBITMAPINFOHEADER,sizeof(TBITMAPINFOHEADER),&read,NULL);

    bytesperline=(width*bpp)/8;
    if ((width*bpp)%8) bytesperline++;

    if (bytesperline%4==0)
        skip=0; else
        skip=4-(bytesperline%4);

    if (bpp<16) {
        DWORD colors=TBITMAPINFOHEADER.biClrUsed;
        if (colors==0) colors=1<<bpp;
        for (x=0;x<colors;x++) {
            ReadFile(handle,&palette[x][2],1,&read,NULL);
            ReadFile(handle,&palette[x][1],1,&read,NULL);
            ReadFile(handle,&palette[x][0],1,&read,NULL);
            ReadFile(handle,&ch,1,&read,NULL);
        }
    }

    SetFilePointer(handle,TBITMAPFILEHEADER.bfOffBits,NULL,FILE_BEGIN);

    bmpbytes=0;
    getbuf=0;
    
    if (TBITMAPINFOHEADER.biCompression==0 && (bpp==1 || bpp==4))  // No compression, bpp=1 or 4
    for (y=height-1;y>=0;y--) {
        for (x=0;x<width;x++) {
            index=bmp_get_bits(handle,bpp);
            ptr[y*width*3+x*3+0]=palette[index][0];
            ptr[y*width*3+x*3+1]=palette[index][1];
            ptr[y*width*3+x*3+2]=palette[index][2];
        }
        for (x=0;x<skip;x++) bmp_get_byte(handle);
        getbuf=0;
    } else
    if (TBITMAPINFOHEADER.biCompression==0 && bpp==8)  // No compression, bpp=8
    for (y=height-1;y>=0;y--) {
        for (x=0;x<width;x++) {
            index=bmp_get_byte(handle);
            ptr[y*width*3+x*3+0]=palette[index][0];
            ptr[y*width*3+x*3+1]=palette[index][1];
            ptr[y*width*3+x*3+2]=palette[index][2];
        }
        for (x=0;x<skip;x++) bmp_get_byte(handle);
    } else
    if (TBITMAPINFOHEADER.biCompression==0 && bpp==24)  // No compression, bpp=24
    for (y=height-1;y>=0;y--) {
        for (x=0;x<width;x++) {
            b=bmp_get_byte(handle);
            g=bmp_get_byte(handle);
            r=bmp_get_byte(handle);
            ptr[y*width*3+x*3+0]=r;
            ptr[y*width*3+x*3+1]=g;
            ptr[y*width*3+x*3+2]=b;

        }
        for (x=0;x<skip;x++) bmp_get_byte(handle);
    } else
    if (TBITMAPINFOHEADER.biCompression==1 && bpp==8) {    // RLE8 compression, bpp=8
        y=height-1;
        x=0;
        do {
            ch=bmp_get_byte(handle);
            ch2=bmp_get_byte(handle);
            
            if (ch==0) {
                if (ch2==0) {   // EOL
                    x=0;
                    y--;
                } else
                if (ch2==1) {   // EOI
                } else
                if (ch2==2) {   // Delta
                    x+=bmp_get_byte(handle);
                    y-=bmp_get_byte(handle);
                } else {        // Absolute mode
                    for (i=0;i<ch2;i++) {
                        ch=bmp_get_byte(handle);
                        ptr[(y*width+x)*3+0]=palette[ch][0];
                        ptr[(y*width+x)*3+1]=palette[ch][1];
                        ptr[(y*width+x)*3+2]=palette[ch][2];
                        x++;
                    }
                    if (ch2%2!=0) bmp_get_byte(handle);
                }
            } else {
                for (i=0;i<ch;i++) {
                    ptr[(y*width+x)*3+0]=palette[ch2][0];
                    ptr[(y*width+x)*3+1]=palette[ch2][1];
                    ptr[(y*width+x)*3+2]=palette[ch2][2];
                    x++;
                }
            }
        } while (ch!=0 || ch2!=1);
    } else
    if (TBITMAPINFOHEADER.biCompression==2 && bpp==4) {    // RLE4 compression, bpp=4
        y=height-1;
        x=0;
        do {
            ch=bmp_get_byte(handle);
            ch2=bmp_get_byte(handle);
            
            if (ch==0) {
                if (ch2==0) {   // EOL
                    x=0;
                    y--;
                } else
                if (ch2==1) {   // EOI
                } else
                if (ch2==2) {   // Delta
                    x+=bmp_get_byte(handle);
                    y-=bmp_get_byte(handle);
                } else {        // Absolute mode
                    for (i=0;i<ch2;i++) {
                        if (i%2==0) ch=bmp_get_byte(handle);
                        ptr[(y*width+x)*3+0]=palette[(ch>>(i%2==0?4:0))&0xf][0];
                        ptr[(y*width+x)*3+1]=palette[(ch>>(i%2==0?4:0))&0xf][1];
                        ptr[(y*width+x)*3+2]=palette[(ch>>(i%2==0?4:0))&0xf][2];
                        x++;
                    }
                    if (ch2%4!=0) bmp_get_byte(handle);
                }
            } else {
                for (i=0;i<ch;i++) {
                    ptr[(y*width+x)*3+0]=palette[(ch2>>(i%2==0?4:0))&0xf][0];
                    ptr[(y*width+x)*3+1]=palette[(ch2>>(i%2==0?4:0))&0xf][1];
                    ptr[(y*width+x)*3+2]=palette[(ch2>>(i%2==0?4:0))&0xf][2];
                    x++;
                }
            }
        } while (ch!=0 || ch2!=1);
    }
}

int BMPinfo(HANDLE handle, WORD *width, WORD *height, WORD *bpp)
{
    DWORD read, size;
    
    SetFilePointer(handle,0,NULL,FILE_BEGIN);
    ReadFile(handle,&TBITMAPFILEHEADER,sizeof(TBITMAPFILEHEADER),&read,NULL);
    ReadFile(handle,&size,4,&read,NULL);
    SetFilePointer(handle,sizeof(TBITMAPFILEHEADER),NULL,FILE_BEGIN);
    if (size==12) {
        ReadFile(handle,&TBITMAPCOREHEADER,sizeof(TBITMAPCOREHEADER),&read,NULL);
        TBITMAPINFOHEADER.biSize=TBITMAPCOREHEADER.bcSize;
        TBITMAPINFOHEADER.biWidth=TBITMAPCOREHEADER.bcWidth;
        TBITMAPINFOHEADER.biHeight=TBITMAPCOREHEADER.bcHeight;
        TBITMAPINFOHEADER.biPlanes=TBITMAPCOREHEADER.bcPlanes;
        TBITMAPINFOHEADER.biBitCount=TBITMAPCOREHEADER.bcBitCount;
        TBITMAPINFOHEADER.biCompression=0;
    } else ReadFile(handle,&TBITMAPINFOHEADER,sizeof(TBITMAPINFOHEADER),&read,NULL);

    if (TBITMAPFILEHEADER.bfTag[0]!='B' || TBITMAPFILEHEADER.bfTag[1]!='M') return(-255);

    // only 1, 4, 8 and 24 bits per pixel allowed for uncompressed
    // only 8 bits per pixel allowed for RLE8
    // only 4 bits per pixel allowed for RLE4
    if (TBITMAPINFOHEADER.biCompression==0 &&
        TBITMAPINFOHEADER.biBitCount!=1 &&
        TBITMAPINFOHEADER.biBitCount!=4 &&
        TBITMAPINFOHEADER.biBitCount!=8 &&
        TBITMAPINFOHEADER.biBitCount!=24) return(-TBITMAPINFOHEADER.biCompression);
    if (TBITMAPINFOHEADER.biCompression!=0 &&
        TBITMAPINFOHEADER.biCompression!=1 &&
        TBITMAPINFOHEADER.biCompression!=2) return(-TBITMAPINFOHEADER.biCompression);
    if (TBITMAPINFOHEADER.biCompression==1 && TBITMAPINFOHEADER.biBitCount!=8) return(-TBITMAPINFOHEADER.biCompression);
    if (TBITMAPINFOHEADER.biCompression==2 && TBITMAPINFOHEADER.biBitCount!=4) return(-TBITMAPINFOHEADER.biCompression);

    if (width!=NULL) *width = abs(TBITMAPINFOHEADER.biWidth);
    if (height!=NULL) *height = abs(TBITMAPINFOHEADER.biHeight);
    if (bpp!=NULL) *bpp = TBITMAPINFOHEADER.biBitCount;
    return(TBITMAPINFOHEADER.biCompression);
}
