//***************************************************************
// This file is part of RPMWCX, a archiver plugin for
// Windows Commander.
// Copyright (C) 2000 Mandryka Yurij  e-mail:braingroup@hotmail.ru
//***************************************************************

//***************************************************************
// Part of code (functions DirectoryExists and ForceDirectories)
// got from Delphi source code
//***************************************************************

//***************************************************************
// This code based on Christian Ghisler (support@ghisler.com) sources
//***************************************************************

{$A-,I-}
unit cpio_io;

interface

uses
  cpio_def;

type
  TStrBuf = array[1..260] of Char;

function  CPIO_ReadHeader(var f : file; var header : CPIO_Header) : Boolean;

function  AlignFilePointer(var f : file; align : Integer) : Boolean;
procedure copy_str2buf(var buf : TStrBuf; s : AnsiString);
procedure BackSlashes(var S : AnsiString);
function  CreateDirectories(Dir : String) : Boolean;
function  correct_filename(oldname : AnsiString) : AnsiString;

implementation

uses
  Windows, SysUtils;

procedure copy_str2buf(var buf : TStrBuf; s : AnsiString);
var
  i_char : Integer;
begin
  FillChar(buf, Sizeof(buf), 0);
  if Length(s) = 0 then Exit;
  if Length(s) > 259 then
    SetLength(s, 259);
  s := s + #0;
  for i_char := 1 to Length(s) do
    buf[i_char] := s[i_char];
end;

procedure BackSlashes(var S : AnsiString);
var i_pos : Integer;
begin
  for i_pos := 1 to Length(S) do
    if S[i_pos] = '/' then S[i_pos] := '\';
end;

function AlignFilePointer;
var
  start : Integer;
  mul   : LongWord;
begin
  Result := False;
  start := FilePos(f);
  case align of
    4 : mul := $FFFFFFFC;
    8 : mul := $FFFFFFF8;
  else
    Exit;
  end;{case}
  if (start mod align) <> 0 then begin
    start := start and mul;
    Inc(start, align);
  end;
  Seek(f, start);
  if IOResult = 0 then Result := True;
end;

function DirectoryExists(const Name: string): Boolean;
var
  Code: Integer;
begin
  Code := GetFileAttributes(PChar(Name));
  Result := (Code <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code <> 0);
end;

function ExcludeTrailingBackslash(Dir:string):string;
begin
  if (length(dir)>0) and (dir[length(dir)]='\') then
    result:=copy(dir,1,length(dir)-1)
  else
    result:=dir;
end;

function CreateDirectories(Dir : String) : Boolean;
begin
  Result := True;
  if Length(Dir) = 0 then
    Result := False
  else begin
    Dir := ExcludeTrailingBackslash(Dir);
    if (Length(Dir) < 3) or DirectoryExists(Dir)
      or (ExtractFilePath(Dir) = Dir) then Exit;
    Result := CreateDirectories(ExtractFilePath(Dir)) and CreateDir(Dir);
  end;
end;

function CPIO_ReadHeader;
var
  tmp_buf    : array [0..259] of Char;
  tmp_str    : String;
  code, i, j : Integer;
  loadlen    : Integer;
  value      : LongWord;
begin
  Result := False;
  for i := 1 to 14 do begin
    if i = 1 then loadlen := 6 else loadlen := 8;
    BlockRead(f, tmp_buf, loadlen);
    if IOResult <> 0 then Exit;
    SetLength(tmp_str, loadlen);
    for j:=1 to loadlen do tmp_str[j] := tmp_buf[j - 1];
    tmp_str := '$' + tmp_str;
    Val(tmp_str, value, code);
    header.records[i] := value;
  end;
  BlockRead(f, tmp_buf, header.records[13]);
  if IOResult <> 0 then Exit;
  SetLength(header.filename, header.records[13] - 1);
  for j:=1 to header.records[13] do header.filename[j] := tmp_buf[j - 1];
  header.origname := header.filename;
  BackSlashes(header.filename);
  if not AlignFilePointer(f, 4) then Exit;

  //Correct file name started with "./" or "/"
  header.filename := correct_filename(header.filename);

  Result := True;
end;

function correct_filename(oldname : AnsiString) : AnsiString;
begin
  Result := oldname;
  if Length(oldname) > 1 then begin
    case oldname[1] of
      '.' :
        case oldname[2] of
          '/', '\' : System.Delete(oldname, 1, 2);
        end;{case}
      '/', '\' : System.Delete(oldname, 1, 1);
    end;{case}
  end;
  Result := oldname;
end;

end.
