// iso_wincmdvc6.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "iso_wincmd.h"
#include <stdlib.h>
#include <assert.h>

// iso_wincmd.cpp : Defines the entry point for the DLL application.
//


typedef struct VolumeDateTime
{
    unsigned char Year;     // Number of years since 1900
    unsigned char Month;    // Month of the year from 1 to 12
    unsigned char Day;      // Day of the Month from 1 to 31
    unsigned char Hour;     // Hour of the day from 0 to 23
    unsigned char Minute;   // Minute of the hour from 0 to 59
    unsigned char Second;   // second of the minute from 0 to 59
             char Zone;     // Offset from Greenwich Mean Time in
                            // number of 15 minute intervals from
                            // -48(West) to +52(East)
} VolumeDateTime;


#define FATTR_HIDDEN       1
#define FATTR_DIRECTORY    2

typedef struct DirectoryRecord
{
    unsigned char LengthOfDirectoryRecord;  //
    unsigned char ExtendedAttributeRecordLength; // Bytes - this field refers to the
                                            // Extended Attribute Record, which provides
                                            // additional information about a file to
                                            // systems that know how to use it. Since
                                            // few systems use it, we will not discuss
                                            // it here. Refer to ISO 9660:1988 for
                                            // more information.
    __int64       LocationOfExtent;         // This is the Logical Block Number of the
                                            // first Logical Block allocated to the file
    __int64       DataLength;               // Length of the file section in bytes
    VolumeDateTime RecordingDateAndTime;    //
    unsigned char FileFlags;                // One Byte, each bit of which is a Flag:
                                            // bit
                                            // 0 File is Hidden if this bit is 1
                                            // 1 Entry is a Directory if this bit is 1
                                            // 2 Entry is an Associated file is this bit is 1
                                            // 3 Information is structured according to
                                            //   the extended attribute record if this
                                            //   bit is 1
                                            // 4 Owner, group and permissions are
                                            //   specified in the extended attribute
                                            //   record if this bit is 1
                                            // 5 Reserved (0)
                                            // 6 Reserved (0)
                                            // 7 File has more than one directory record
                                            //   if this bit is 1
    unsigned char FileUnitSize;             // This field is only valid if the file is
                                            // recorded in interleave mode
                                            // Otherwise this field is (00)
    unsigned char InterleaveGapSize;        // This field is only valid if the file is
                                            // recorded in interleave mode
                                            // Otherwise this field is (00)
    int VolumeSequenceNumber;               // The ordinal number of the volume in the Volume
                                            // Set on which the file described by the
                                            // directory record is recorded.
    unsigned char LengthOfFileIdentifier;   // Length of File Identifier (LEN_FI)
    unsigned char FileIdentifier[1];
} DirectoryRecord;

typedef struct PathTableRecord
{
    unsigned char LengthOfDirectoryIdentifier;  // Length of Directory Identifier (LEN_DI)
    unsigned char ExtendedAttributeRecordLength;// If an Extended Attribute Record is
                                                // recorded, this is the length in Bytes.
                                                // Otherwise, this is (00)
    unsigned int  LocationOfExtent;             // Logical Block Number of the first Logical
                                                // Block allocated to the Directory
    unsigned short int ParentDirectoryNumber;   // The record number in the Path Table for
                                                // the parent directory of this directory
    unsigned char DirectoryIdentifier[1];       // This field is the same as in the Directory
                                                // Record
} PathTableRecord;


typedef struct PrimaryVolumeDescriptor
{
    unsigned char VolumeDescriptorType;
    unsigned char StandardIdentifier[5];        // CD001
    unsigned char VolumeDescriptorVersion;
    unsigned char Ununsed;
    unsigned char SystemIdentifier[32];
    unsigned char VolumeIdentifier[32];
    unsigned char Unused2[8];
    __int64       VolumeSpaceSize;              // Number of logical blocks in the Volume
    unsigned char Unused3[32];
    unsigned int  VolumeSetSize;                // The assigned Volume Set size of the Volume
    unsigned int  VolumeSequenceNumber;         // The ordinal number of the volume in
                                                // the Volume Set
    unsigned int  LogicalBlockSize;             // The size in bytes of a Logical Block
    __int64       PathTableSize;                // Length in bytes of the path table
    unsigned int  LocationOfTypeLPathTable;     // Logical Block Number of first Block allocated
                                                // to the Type L Path Table
    unsigned int  LocationOfOptionalTypeLPathTable; // 0 if Optional Path Table was not recorded,
                                                // otherwise, Logical Block Number of first
                                                // Block allocated to the Optional Type L
                                                // Path Table
    unsigned int  LocationOfTypeMPathTable;     // Logical Block Number of first Block
                                                // allocated to the Type M
    unsigned int  LocationOfOptionalTypeMPathTable; // 0 if Optional Path Table was not
                                                // recorded, otherwise, Logical Path Table,
                                                // Block Number of first Block allocated to the
                                                // Type M Path Table.
    DirectoryRecord DirectoryRecordForRootDirectory; // This is the actual directory record for
                                                // the top of the directory structure
    unsigned char VolumeSetIdentifier[128];     // Name of the multiple volume set of which
                                                // this volume is a member.
    unsigned char PublisherIdentifier[128];     // Identifies who provided the actual data
                                                // contained in the files. acharacters allowed.
    unsigned char DataPreparerIdentifier[128];  // Identifies who performed the actual
                                                // creation of the current volume.
    unsigned char ApplicationIdentifier[128];   // Identifies the specification of how the
                                                // data in the files are recorded.
    unsigned char CopyrightFileIdentifier[37];  // Identifies the file in the root directory
                                                // that contains the copyright notice for
                                                // this volume
    unsigned char AbstractFileIdentifier[37];   // Identifies the file in the root directory
                                                // that contains the abstract statement for
                                                // this volume
    unsigned char BibliographicFileIdentifier[37]; // Identifies the file in the root directory
                                                // that contains bibliographic records.
    unsigned char VolumeCreationDateAndTime[17];// Date and time at which the volume was created
    unsigned char VolumeModificationDateAndTime[17]; // Date and time at which the volume was
                                                // last modified
    unsigned char VolumeExpirationDateAndTime[17]; // Date and Time at which the information in
                                                // the volume may be considered obsolete.
    unsigned char VolumeEffectiveDateAndTime[17];// Date and Time at which the information
                                                // in the volume may be used
    unsigned char FileStructureVersion;         // 1
    unsigned char ReservedForFutureStandardization; // 0
    unsigned char ApplicationUse[512];          // This field is reserved for application use.
                                                // Its content is not specified by ISO-9660.
    unsigned char ReservedForFutureStandardization2[653]; // 0
} PrimaryVolumeDescriptor; // sizeof( PrimaryVolumeDescriptor ) must be 2048


class CIsoFile : public DirectoryRecord
{
public:
    typedef CArray< CIsoFile, CIsoFile& > t_Files;

    t_Files  m_Files;
    CString  m_Name;
    int      m_Length;
    FILETIME m_WriteTime;

    CIsoFile() { m_Length = 0; }
    CIsoFile( const DirectoryRecord* file, bool unicode = false );
    CIsoFile( const CIsoFile& file ) { *this = file; }
    void operator =( const CIsoFile& file );
};

#define ROTATE(a) ((a) >> 8) + ((a) << 8)

CIsoFile::CIsoFile( const DirectoryRecord* file, bool unicode )
{
    assert( file );

    DirectoryRecord::operator =( *file );
    if( !unicode )
    {
        char name[1000];
        strncpy( name, (const char*)file->FileIdentifier, file->LengthOfFileIdentifier );
        name[file->LengthOfFileIdentifier] = 0;
        if( strrchr( name, ';' ) )
            *strrchr( name, ';' ) = 0;
        if( strlen( name ) && name[strlen(name) - 1] == '.' )
            name[strlen(name) - 1] = 0;
        m_Name = name;
        m_Length = file->LengthOfDirectoryRecord;
    }
    else
    {
        wchar_t name[1000];
        for( int i = 0; i < file->LengthOfFileIdentifier / 2; i++ )
            name[i] = ROTATE( ((wchar_t*)file->FileIdentifier)[i] );
        name[file->LengthOfFileIdentifier / 2] = 0;
        if( wcsrchr( name, ';' ) )
            *wcsrchr( name, ';' ) = 0;
        if( wcslen( name ) && name[wcslen(name) - 1] == '.' )
            name[wcslen(name) - 1] = 0;
        m_Name = name;
        m_Length = file->LengthOfDirectoryRecord;
    }

    SYSTEMTIME time;
    memset( &time, 0, sizeof( time ) );
    time.wYear   = file->RecordingDateAndTime.Year + 1900;
    time.wMonth  = file->RecordingDateAndTime.Month;
    time.wDay    = file->RecordingDateAndTime.Day;
    time.wHour   = file->RecordingDateAndTime.Hour;
    time.wMinute = file->RecordingDateAndTime.Minute;
    time.wSecond = file->RecordingDateAndTime.Second;

    SystemTimeToFileTime( &time, &m_WriteTime );
}

void CIsoFile::operator =( const CIsoFile& file )
{
    DirectoryRecord::operator =( file );

    m_Files.RemoveAll();
    m_Files.Append( file.m_Files );
    m_Name = file.m_Name;
    m_Length = file.m_Length;
    m_WriteTime = file.m_WriteTime;
}

class CIsoDir : public PathTableRecord
{
public:
    CString m_Name;
    int     m_Length;

    CIsoDir() { m_Length = 0; }
    CIsoDir( const PathTableRecord* dir, bool unicode = false );
    void operator =( const CIsoDir& dir );
};

CIsoDir::CIsoDir( const PathTableRecord* dir, bool unicode )
{
    assert( dir );
    PathTableRecord::operator =( *dir );
    if( !unicode )
    {
        char name[1000];
        strncpy( name, (const char*)dir->DirectoryIdentifier, dir->LengthOfDirectoryIdentifier );
        name[dir->LengthOfDirectoryIdentifier] = 0;
        m_Name = name;
    }
    else
    {
        wchar_t name[1000];
        for( int i = 0; i < dir->LengthOfDirectoryIdentifier / 2; i++ )
            name[i] = ROTATE( ((wchar_t*)dir->DirectoryIdentifier)[i] );
        name[dir->LengthOfDirectoryIdentifier / 2] = 0;
        m_Name = name;
    }
    m_Length = (dir->LengthOfDirectoryIdentifier + 9) & ~1;
}

void CIsoDir::operator =( const CIsoDir& dir )
{
    m_Name = dir.m_Name;
    m_Length = dir.m_Length;
}

class CIsoImage
{
public:
    typedef CArray<CIsoDir, CIsoDir&>    t_PathTable;
    typedef CArray<CIsoFile*, CIsoFile*> t_Path;

    CIsoFile                m_File;
    CIsoFile                m_FileUnicode;
    HANDLE                  m_Image;
    PrimaryVolumeDescriptor m_Pvd;
    PrimaryVolumeDescriptor m_PvdUnicode;
    t_PathTable             m_PathTable;
    t_PathTable             m_PathTableUnicode;
    t_Path                  m_Path;
    bool                    m_Unicode;
    CString                 m_CurrentDir;
    CString                 m_CurrentPath;
    CString                 m_Title;
    CString                 m_CurrentDirOem;
    char                    m_CurrentDirChar[1000];
    char                    m_TitleChar[1000];
    int                     m_Position;
    CIsoFile::t_Files       m_FileList;
    tProcessDataProc        m_ProcessDataProc;
    DWORD                   m_OffsetIntoFile;           // offset into the file to the start of ISO image (WinOnCD .c2d)

    CIsoImage() {m_Image = 0; m_Path.Add( &m_File ); m_Position = 0; m_OffsetIntoFile = 0;
                 m_ProcessDataProc = 0;}
    CIsoImage( const char* name );
    ~CIsoImage() { if( m_Image != INVALID_HANDLE_VALUE ) CloseHandle( m_Image ); }

    DWORD ReadData( DWORD block, DWORD size, void* data );
    DWORD ReadDataByPos( DWORD position, DWORD size, void* data );
    DWORD Seek( DWORD position, int mode );
    bool LoadTree();
    bool ReadFolder( CIsoFile::t_Files& files, DWORD block, DWORD size, bool unicode );
    bool LoadPathTable();
    bool CD( const char* dir );
    CIsoFile* FindFile( const char* name, const CIsoFile::t_Files& dir );
    void GetCurrentPath();
    bool ExtractFile( const char* name, const char* dest, const CIsoFile::t_Files& directory );
    void SplitPath( const char* path, CArray< CString, CString& >& names );
    bool CreateFileList( const char* root, CIsoFile::t_Files& files );
};

CIsoImage::CIsoImage( const char* name )
{
    m_Position = 0;
    m_OffsetIntoFile = 0;
    m_ProcessDataProc = 0;
    m_File.m_Name = name;

    m_Image = CreateFile( name, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0 );
    if( m_Image == INVALID_HANDLE_VALUE )
        return;

    // First check to see if we're dealing with a WinOnCD "image" file or a normal ISO image
    {
        char	idstr[5];

        if( ReadDataByPos(0x8001, 5, (void*) idstr) != 5 )
        {
            CloseHandle( m_Image );
            m_Image = INVALID_HANDLE_VALUE;
            return;
        }
        if( memcmp( &idstr[0], "CD001", 5 ) != 0 )
        {
            // It's not a normal ISO file, lets assume it's a c2d (might need further testing to be sure)
            m_OffsetIntoFile = 0x20000;
        }
    }

    if( Seek( 0x8000, FILE_BEGIN ) != 0x8000 )
    {
        CloseHandle( m_Image );
        m_Image = INVALID_HANDLE_VALUE;
        return;
    }

    m_File.m_Name.OemToAnsi();

    assert( sizeof( m_Pvd ) == 0x800 );

    DWORD read;
    ReadFile( m_Image, &m_Pvd, sizeof( m_Pvd ), &read, 0 ); // read standard descriptor
    if( read < sizeof( m_Pvd ) )
    {
        CloseHandle( m_Image );
        m_Image = INVALID_HANDLE_VALUE;
        return;
    }

    if( memcmp( &m_Pvd.StandardIdentifier, "CD001", 5 ) )
    {
        CloseHandle( m_Image );
        m_Image = INVALID_HANDLE_VALUE;
        return;
    }

    m_Unicode = false;
    while( !m_Unicode )
    {
        ReadFile( m_Image, &m_PvdUnicode, sizeof( m_PvdUnicode ), &read, 0 );
                                                            // read unicode descriptor
        if( read < sizeof( m_PvdUnicode ) )
        {
            CloseHandle( m_Image );
            m_Image = INVALID_HANDLE_VALUE;
            return;
        }

        if( memcmp( &m_PvdUnicode.StandardIdentifier, "CD001", 5 ) )
            break;
        if( m_PvdUnicode.VolumeDescriptorType != 2                                ||
            !(DWORD)m_PvdUnicode.DirectoryRecordForRootDirectory.LocationOfExtent ||
            !(DWORD)m_PvdUnicode.DirectoryRecordForRootDirectory.DataLength )
            m_Unicode = false;
        else
            m_Unicode = true;
    }

    m_Path.Add( m_Unicode ? &m_FileUnicode : &m_File );
}

DWORD CIsoImage::Seek( DWORD position, int mode )
{
	DWORD	rc;

    if (mode == FILE_BEGIN)
    {
        rc = SetFilePointer( m_Image, (LONG)position + m_OffsetIntoFile, 0, mode );
        return (rc > m_OffsetIntoFile ? rc - m_OffsetIntoFile : 0);
    }
    else
        return SetFilePointer( m_Image, (LONG)position, 0, mode );
}

DWORD CIsoImage::ReadData( DWORD block, DWORD size, void* data )
{
    DWORD position = block * (short int)m_Pvd.LogicalBlockSize;
    if( Seek( position, FILE_BEGIN ) != position )
        return 0;

    DWORD read;

    if( ReadFile( m_Image, data, size, &read, 0 ) )
        return read;
    else
        return 0;
}

DWORD CIsoImage::ReadDataByPos( DWORD position, DWORD size, void* data )
{
    if( Seek( position, FILE_BEGIN ) != position )
        return 0;

    DWORD read;

    if( ReadFile( m_Image, data, size, &read, 0 ) )
        return read;
    else
        return 0;
}

bool CIsoImage::ReadFolder( CIsoFile::t_Files& files, DWORD block, DWORD size, bool unicode )
{
    assert( block );
    assert( size );
    unsigned char* data = (unsigned char*)malloc( size );
    if( !ReadData( block, size, data ) )
    {
        delete data;
        return false;
    }

    for( DWORD i = 0, j = 0; i < size - 0x22; j++ )
    {
        while( !data[i] && i < size - 0x22 ) i++;
        if( i >= size - 0x22 )
            break;
        CIsoFile file( (DirectoryRecord*)(data + i), unicode );
        if( !file.m_Length )
            break;
        if( j > 1 )
        {
            if( file.FileFlags & FATTR_DIRECTORY )
                ReadFolder( file.m_Files, (DWORD)file.LocationOfExtent,
                            (DWORD)file.DataLength, unicode );
            if( file.m_Name.GetLength() )
                files.Add( file );
        }
        i += file.m_Length;
    }

    delete data;
    return true;
}

bool CIsoImage::LoadTree()
{
    if( ReadFolder( m_File.m_Files,
                    (DWORD)m_Pvd.DirectoryRecordForRootDirectory.LocationOfExtent,
                    (DWORD)m_Pvd.DirectoryRecordForRootDirectory.DataLength, false ) &&
        !m_Unicode )
        return true;
    else
        return ReadFolder( m_FileUnicode.m_Files,
                    (DWORD)m_PvdUnicode.DirectoryRecordForRootDirectory.LocationOfExtent,
                    (DWORD)m_PvdUnicode.DirectoryRecordForRootDirectory.DataLength, true );
}

bool CIsoImage::LoadPathTable()
{
    DWORD size = (DWORD)m_Pvd.PathTableSize;
    DWORD block = (DWORD)m_Pvd.LocationOfTypeLPathTable;

    if( !block || !size )
    {
        assert( 0 );
        return false;
    }

    unsigned char* data = (unsigned char*)malloc( size );
    if( !ReadData( block, size, data ) )
    {
        delete data;
        return false;
    }

    DWORD i;

    for( i = 0; i < size; )
    {
        CIsoDir dir( (PathTableRecord*)(data + i) );
        if( !dir.m_Length )
            break;
        m_PathTable.Add( dir );
        i += dir.m_Length;
    }

    delete data;
    if( !m_Unicode )
        return true;

    size = (DWORD)m_PvdUnicode.PathTableSize;
    block = (DWORD)m_PvdUnicode.LocationOfTypeLPathTable;

    if( !block || !size )
    {
        assert( 0 );
        return false;
    }

    data = (unsigned char*)malloc( size );
    if( !ReadData( block, size, data ) )
    {
        delete data;
        return false;
    }

    for( i = 0; i < size; )
    {
        CIsoDir dir( (PathTableRecord*)(data + i), true );
        if( !dir.m_Length )
            break;
        m_PathTableUnicode.Add( dir );
        i += dir.m_Length;
    }

    delete data;

    return true;
}

void CIsoImage::GetCurrentPath()
{
    m_CurrentPath = "";
    for( int i = 1; i < m_Path.GetSize(); i++ )
    {
        if( m_CurrentPath.GetLength() )
            m_CurrentPath += "\\";
        m_CurrentPath += m_Path[i]->m_Name;
    }
}

CIsoFile* CIsoImage::FindFile( const char* name, const CIsoFile::t_Files& dir )
{
    CArray< CString, CString& > names;
    SplitPath( name, names );

    CIsoFile::t_Files* cur_dir = (CIsoFile::t_Files*)&dir;

    CIsoFile* file = 0;

    for( int j = 0; j < names.GetSize(); j++ )
    {
        file = 0;
        for( int i = 0; i < cur_dir->GetSize(); i++ )
        {
            CIsoFile* f = &((*cur_dir)[i]);
            if( !CString( names[j] ).CompareNoCase( f->m_Name ) )
            {
                cur_dir = &(f->m_Files);
                file = f;
                break;
            }
        }
        if( !file )
            return 0;
    }
    return file;
}

void CIsoImage::SplitPath( const char* path, CArray< CString, CString& >& names )
{
    while( *path == '\\' || *path == '/' ) path++;

    char d_[1000];
    char* begin = d_;
    char* end = begin;
    strcpy( d_, path );
    char* last = begin + strlen( path );

    while( begin < last )
    {
        while( *end != '\\' && *end != '/' && *end ) end++;
        *end = 0;
        CString str( begin );
        if( str.GetLength() )
            names.Add( str );
        begin = end + 1;
        while( begin < last && *begin && (*begin == '\\' || *begin == '/') ) begin++;
        end = begin;
    }
}

bool CIsoImage::CD( const char* dir )
{
    if( !dir[0] )
        return true;

    // cd ..
    if( !strcmp( dir, ".." ) )
    {
        if( m_Path.GetSize() <= 2 )
        {
            m_Path.RemoveAll();
            m_Path.Add( m_Unicode ? &m_FileUnicode : &m_File );
            m_CurrentDir = "";
            m_CurrentDirOem = "";
            m_CurrentPath = "";
            return true;
        }
        else
        {
            m_Path.RemoveAt( m_Path.GetSize() - 1 );
            m_CurrentDir = m_Path[m_Path.GetSize() - 1]->m_Name;
            m_CurrentDirOem = m_CurrentDir;
            m_CurrentDirOem.AnsiToOem();
            GetCurrentPath();
            return true;
        }
    }

    // cd /
    if( !strcmp( dir, "\\" ) || !strcmp( dir, "/" ) )
    {
        m_Path.RemoveAll();
        m_Path.Add( m_Unicode ? &m_FileUnicode : &m_File );
        m_CurrentDir = "";
        m_CurrentDirOem = "";
        m_CurrentPath = "";
        return true;
    }

    // path from root folder ?
    if( *dir == '\\' || *dir == '/' )
    {
        CD( "\\" );
        while( *dir == '\\' || *dir == '/' ) dir++;
    }

    CArray< CString, CString& > path;
    SplitPath( dir, path );

    for( int i = 0; i < path.GetSize(); i++ )
    {
        CIsoFile* file = FindFile( path[i], m_Path[m_Path.GetSize() - 1]->m_Files );
        if( !file )
            break;

        m_Path.Add( file );
    }

    m_CurrentDir = m_Path[m_Path.GetSize() - 1]->m_Name;
    m_CurrentDirOem = m_CurrentDir;
    m_CurrentDirOem.AnsiToOem();
    GetCurrentPath();

    return true;
}

bool CIsoImage::ExtractFile( const char* name, const char* dest,
                             const CIsoFile::t_Files& directory )
{
    CArray< CString, CString& > path;
	int ok=1;
    SplitPath( name, path );

    CIsoFile::t_Files* dir;
    if( *name == '/' || *name == '\\' )
        dir = m_Unicode ? &m_FileUnicode.m_Files : &m_File.m_Files;
    else
        dir = (CIsoFile::t_Files*)&directory;

    CIsoFile* file = FindFile( name, *dir );
    if( !file )
        return false;

    if( file->FileFlags & FATTR_DIRECTORY )
    {
        CString name = CString( dest ) + "\\" + file->m_Name;
        name.AnsiToOem();

        if( CreateDirectory( name, 0 ) )
            for( int i = 0; i < file->m_Files.GetSize(); i++ )
                ExtractFile( file->m_Files[i].m_Name,
                             CString( dest ) + "\\" + file->m_Name,
                             file->m_Files );
        else
            return false;
    }
    else
    {
#define READ_SIZE 100000
        DWORD flags = FILE_ATTRIBUTE_ARCHIVE |
                      (file->FileFlags & FATTR_HIDDEN ? FILE_ATTRIBUTE_HIDDEN : 0);
        CString name = CString( dest ) + "\\" + file->m_Name;
        name.AnsiToOem();
        HANDLE hfile = CreateFile( dest, GENERIC_WRITE, FILE_SHARE_READ, 0,
                                         CREATE_ALWAYS, flags, 0 );
        if( hfile == INVALID_HANDLE_VALUE )
            return false;

        void* buffer= malloc( READ_SIZE );

        DWORD position = (DWORD)file->LocationOfExtent * (m_Unicode ?
                                        (short int)m_PvdUnicode.LogicalBlockSize :
                                        (short int)m_Pvd.LogicalBlockSize);
        for( DWORD i = 0; i < (DWORD)file->DataLength; i += READ_SIZE )
        {
            DWORD write;
            DWORD read = ReadDataByPos( position + i,
                                        min( READ_SIZE, (DWORD)file->DataLength - i),
                                        buffer );
            WriteFile( hfile, buffer, read, &write, 0 );
            if( m_ProcessDataProc ) {
                ok=m_ProcessDataProc( file->m_Name, read );
				if (ok==0)
					break;
			}
        }
        SetFileTime( hfile, &file->m_WriteTime, &file->m_WriteTime, &file->m_WriteTime );
        CloseHandle( hfile );

        delete buffer;
    }

    return (ok!=0);
}

bool CIsoImage::CreateFileList( const char* root, CIsoFile::t_Files& files )
{
    CString path = root;
    for( int i = 0; i < files.GetSize(); i++ )
    {
        CIsoFile file = files[i];
        file.m_Name = CString( root ) + (strlen( root ) ? "\\" : "" ) + file.m_Name;
        m_FileList.Add( file );
        if( file.FileFlags & FATTR_DIRECTORY )
            CreateFileList( file.m_Name, files[i].m_Files );
    }
    return true;
}

ISO_WINCMD_API HANDLE CALLDECL OpenArchive( tOpenArchiveData* ArchiveData )
{
    ArchiveData->CmtBuf     = 0;
    ArchiveData->CmtBufSize = 0;
    ArchiveData->CmtSize    = 0;
    ArchiveData->CmtState   = 0;

    CIsoImage* arc = new CIsoImage( ArchiveData->ArcName );
    if( !arc->m_Image || arc->m_Image == INVALID_HANDLE_VALUE )
    {
        delete arc;
        ArchiveData->OpenResult = E_EOPEN;
        return 0;
    }
    else
    {
        ArchiveData->OpenResult = 0;
        arc->LoadTree();
        arc->m_FileList.RemoveAll();
        arc->CreateFileList( "", arc->m_Unicode ? arc->m_FileUnicode.m_Files :
                                                  arc->m_File.m_Files );
        return arc;
    }
}

ISO_WINCMD_API int CALLDECL ReadHeader( HANDLE hArcData, tHeaderData* HeaderData )
{
    CIsoImage* arc = (CIsoImage*)hArcData;

    if( arc->m_Position >= arc->m_FileList.GetSize() )
    {
        arc->m_Position = 0;
        return E_END_ARCHIVE;
    }

    strcpy( HeaderData->ArcName, arc->m_File.m_Name );
    strcpy( HeaderData->FileName,
            arc->m_FileList[arc->m_Position].m_Name );

    HeaderData->FileAttr = FA_ARCHIVE;
    HeaderData->FileAttr |= arc->m_FileList[arc->m_Position].FileFlags & FATTR_DIRECTORY ?
                            FA_DIRECTORY : 0;
    HeaderData->FileAttr |= arc->m_FileList[arc->m_Position].FileFlags & FATTR_HIDDEN ?
                            FA_HIDDEN : 0;

    WORD time;
    WORD date;
    FileTimeToDosDateTime( &arc->m_FileList[arc->m_Position].m_WriteTime,
                           &date, &time );
    HeaderData->FileTime = (((DWORD)date) << 16) + time;
    HeaderData->PackSize = HeaderData->UnpSize =
        (DWORD)arc->m_FileList[arc->m_Position].DataLength;

    arc->m_Position++;

    return 0;
}

ISO_WINCMD_API int CALLDECL ProcessFile( HANDLE hArcData, int Operation,
                                         char *DestPath, char *DestName )
{
    CIsoImage* arc = (CIsoImage*)hArcData;

    assert( arc->m_Position );

    CString n = arc->m_FileList[arc->m_Position - 1].m_Name;
    if( Operation == PK_SKIP || Operation == PK_TEST ) return 0;

    char dest[1000] = "";

    if( DestPath )
        strcpy( dest, DestPath );
    if( DestName )
        strcat( dest, DestName );

    return !arc->ExtractFile( arc->m_FileList[arc->m_Position - 1].m_Name,
                              dest, arc->m_Unicode ? arc->m_FileUnicode.m_Files :
                                                     arc->m_File.m_Files );
}

ISO_WINCMD_API int CALLDECL CloseArchive (HANDLE hArcData)
{
    CIsoImage* arc = (CIsoImage*)hArcData;
    delete arc;
    return 0;
}

ISO_WINCMD_API void CALLDECL SetChangeVolProc( HANDLE, tChangeVolProc )
{
}

ISO_WINCMD_API void CALLDECL SetProcessDataProc( HANDLE hArcData,
                                                 tProcessDataProc pProcessDataProc)
{
    CIsoImage* arc = (CIsoImage*)hArcData;

    arc->m_ProcessDataProc = pProcessDataProc;
}


