/*
 * This file is part of WCX_SFF. A Windows Commander plugin for SFF files.
 * Copyright (C) 2001  Amenna Heratsini
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * email: amennah@hotmail.com
 */

#include <windows.h>
#include "sff.h"

#pragma pack(push,1)
static struct {
    DWORD SFF_Id;
    BYTE Version;
    BYTE reserved;
    WORD UserInformation;
    WORD PageCount;
    WORD OffsetFirstPageHeader;
    DWORD OffsetLastPageHeader;
    DWORD OffsetDocumentEnd;
} SFF_HEADER;

static struct {
    BYTE PageHeaderId;
    BYTE PageHeaderLen;
    BYTE ResolutionVertical;
    BYTE ResolutionHorizontal;
    BYTE Coding;
    BYTE reserved;
    WORD LineLength;
    WORD PageLength;
    DWORD OffsetPreviousPage;
    DWORD OffsetNextPage;
} SFF_PAGE;
#pragma pack(pop)

typedef struct tableentry {
    unsigned short length;      /* bit length of g3 code */
    unsigned short code;        /* g3 code */
    short          runlen;      /* run length in bits */
} tableentry;

const tableentry Group3FaxWhiteCodes[104] = {
    { 8, 0x35, 0 },     /* 0011 0101 */
    { 6, 0x7, 1 },      /* 0001 11 */
    { 4, 0x7, 2 },      /* 0111 */
    { 4, 0x8, 3 },      /* 1000 */
    { 4, 0xB, 4 },      /* 1011 */
    { 4, 0xC, 5 },      /* 1100 */
    { 4, 0xE, 6 },      /* 1110 */
    { 4, 0xF, 7 },      /* 1111 */
    { 5, 0x13, 8 },     /* 1001 1 */
    { 5, 0x14, 9 },     /* 1010 0 */
    { 5, 0x7, 10 },     /* 0011 1 */
    { 5, 0x8, 11 },     /* 0100 0 */
    { 6, 0x8, 12 },     /* 0010 00 */
    { 6, 0x3, 13 },     /* 0000 11 */
    { 6, 0x34, 14 },    /* 1101 00 */
    { 6, 0x35, 15 },    /* 1101 01 */
    { 6, 0x2A, 16 },    /* 1010 10 */
    { 6, 0x2B, 17 },    /* 1010 11 */
    { 7, 0x27, 18 },    /* 0100 111 */
    { 7, 0xC, 19 },     /* 0001 100 */
    { 7, 0x8, 20 },     /* 0001 000 */
    { 7, 0x17, 21 },    /* 0010 111 */
    { 7, 0x3, 22 },     /* 0000 011 */
    { 7, 0x4, 23 },     /* 0000 100 */
    { 7, 0x28, 24 },    /* 0101 000 */
    { 7, 0x2B, 25 },    /* 0101 011 */
    { 7, 0x13, 26 },    /* 0010 011 */
    { 7, 0x24, 27 },    /* 0100 100 */
    { 7, 0x18, 28 },    /* 0011 000 */
    { 8, 0x2, 29 },     /* 0000 0010 */
    { 8, 0x3, 30 },     /* 0000 0011 */
    { 8, 0x1A, 31 },    /* 0001 1010 */
    { 8, 0x1B, 32 },    /* 0001 1011 */
    { 8, 0x12, 33 },    /* 0001 0010 */
    { 8, 0x13, 34 },    /* 0001 0011 */
    { 8, 0x14, 35 },    /* 0001 0100 */
    { 8, 0x15, 36 },    /* 0001 0101 */
    { 8, 0x16, 37 },    /* 0001 0110 */
    { 8, 0x17, 38 },    /* 0001 0111 */
    { 8, 0x28, 39 },    /* 0010 1000 */
    { 8, 0x29, 40 },    /* 0010 1001 */
    { 8, 0x2A, 41 },    /* 0010 1010 */
    { 8, 0x2B, 42 },    /* 0010 1011 */
    { 8, 0x2C, 43 },    /* 0010 1100 */
    { 8, 0x2D, 44 },    /* 0010 1101 */
    { 8, 0x4, 45 },     /* 0000 0100 */
    { 8, 0x5, 46 },     /* 0000 0101 */
    { 8, 0xA, 47 },     /* 0000 1010 */
    { 8, 0xB, 48 },     /* 0000 1011 */
    { 8, 0x52, 49 },    /* 0101 0010 */
    { 8, 0x53, 50 },    /* 0101 0011 */
    { 8, 0x54, 51 },    /* 0101 0100 */
    { 8, 0x55, 52 },    /* 0101 0101 */
    { 8, 0x24, 53 },    /* 0010 0100 */
    { 8, 0x25, 54 },    /* 0010 0101 */
    { 8, 0x58, 55 },    /* 0101 1000 */
    { 8, 0x59, 56 },    /* 0101 1001 */
    { 8, 0x5A, 57 },    /* 0101 1010 */
    { 8, 0x5B, 58 },    /* 0101 1011 */
    { 8, 0x4A, 59 },    /* 0100 1010 */
    { 8, 0x4B, 60 },    /* 0100 1011 */
    { 8, 0x32, 61 },    /* 0011 0010 */
    { 8, 0x33, 62 },    /* 0011 0011 */
    { 8, 0x34, 63 },    /* 0011 0100 */
    { 5, 0x1B, 64 },    /* 1101 1 */
    { 5, 0x12, 128 },   /* 1001 0 */
    { 6, 0x17, 192 },   /* 0101 11 */
    { 7, 0x37, 256 },   /* 0110 111 */
    { 8, 0x36, 320 },   /* 0011 0110 */
    { 8, 0x37, 384 },   /* 0011 0111 */
    { 8, 0x64, 448 },   /* 0110 0100 */
    { 8, 0x65, 512 },   /* 0110 0101 */
    { 8, 0x68, 576 },   /* 0110 1000 */
    { 8, 0x67, 640 },   /* 0110 0111 */
    { 9, 0xCC, 704 },   /* 0110 0110 0 */
    { 9, 0xCD, 768 },   /* 0110 0110 1 */
    { 9, 0xD2, 832 },   /* 0110 1001 0 */
    { 9, 0xD3, 896 },   /* 0110 1001 1 */
    { 9, 0xD4, 960 },   /* 0110 1010 0 */
    { 9, 0xD5, 1024 },  /* 0110 1010 1 */
    { 9, 0xD6, 1088 },  /* 0110 1011 0 */
    { 9, 0xD7, 1152 },  /* 0110 1011 1 */
    { 9, 0xD8, 1216 },  /* 0110 1100 0 */
    { 9, 0xD9, 1280 },  /* 0110 1100 1 */
    { 9, 0xDA, 1344 },  /* 0110 1101 0 */
    { 9, 0xDB, 1408 },  /* 0110 1101 1 */
    { 9, 0x98, 1472 },  /* 0100 1100 0 */
    { 9, 0x99, 1536 },  /* 0100 1100 1 */
    { 9, 0x9A, 1600 },  /* 0100 1101 0 */
    { 6, 0x18, 1664 },  /* 0110 00 */
    { 9, 0x9B, 1728 },  /* 0100 1101 1 */
    { 11, 0x8, 1792 },  /* 0000 0001 000 */
    { 11, 0xC, 1856 },  /* 0000 0001 100 */
    { 11, 0xD, 1920 },  /* 0000 0001 101 */
    { 12, 0x12, 1984 }, /* 0000 0001 0010 */
    { 12, 0x13, 2048 }, /* 0000 0001 0011 */
    { 12, 0x14, 2112 }, /* 0000 0001 0100 */
    { 12, 0x15, 2176 }, /* 0000 0001 0101 */
    { 12, 0x16, 2240 }, /* 0000 0001 0110 */
    { 12, 0x17, 2304 }, /* 0000 0001 0111 */
    { 12, 0x1C, 2368 }, /* 0000 0001 1100 */
    { 12, 0x1D, 2432 }, /* 0000 0001 1101 */
    { 12, 0x1E, 2496 }, /* 0000 0001 1110 */
    { 12, 0x1F, 2560 }}; /* 0000 0001 1111 */

const tableentry Group3FaxBlackCodes[104] = {
    { 10, 0x37, 0 },    /* 0000 1101 11 */
    { 3, 0x2, 1 },      /* 010 */
    { 2, 0x3, 2 },      /* 11 */
    { 2, 0x2, 3 },      /* 10 */
    { 3, 0x3, 4 },      /* 011 */
    { 4, 0x3, 5 },      /* 0011 */
    { 4, 0x2, 6 },      /* 0010 */
    { 5, 0x3, 7 },      /* 0001 1 */
    { 6, 0x5, 8 },      /* 0001 01 */
    { 6, 0x4, 9 },      /* 0001 00 */
    { 7, 0x4, 10 },     /* 0000 100 */
    { 7, 0x5, 11 },     /* 0000 101 */
    { 7, 0x7, 12 },     /* 0000 111 */
    { 8, 0x4, 13 },     /* 0000 0100 */
    { 8, 0x7, 14 },     /* 0000 0111 */
    { 9, 0x18, 15 },    /* 0000 1100 0 */
    { 10, 0x17, 16 },   /* 0000 0101 11 */
    { 10, 0x18, 17 },   /* 0000 0110 00 */
    { 10, 0x8, 18 },    /* 0000 0010 00 */
    { 11, 0x67, 19 },   /* 0000 1100 111 */
    { 11, 0x68, 20 },   /* 0000 1101 000 */
    { 11, 0x6C, 21 },   /* 0000 1101 100 */
    { 11, 0x37, 22 },   /* 0000 0110 111 */
    { 11, 0x28, 23 },   /* 0000 0101 000 */
    { 11, 0x17, 24 },   /* 0000 0010 111 */
    { 11, 0x18, 25 },   /* 0000 0011 000 */
    { 12, 0xCA, 26 },   /* 0000 1100 1010 */
    { 12, 0xCB, 27 },   /* 0000 1100 1011 */
    { 12, 0xCC, 28 },   /* 0000 1100 1100 */
    { 12, 0xCD, 29 },   /* 0000 1100 1101 */
    { 12, 0x68, 30 },   /* 0000 0110 1000 */
    { 12, 0x69, 31 },   /* 0000 0110 1001 */
    { 12, 0x6A, 32 },   /* 0000 0110 1010 */
    { 12, 0x6B, 33 },   /* 0000 0110 1011 */
    { 12, 0xD2, 34 },   /* 0000 1101 0010 */
    { 12, 0xD3, 35 },   /* 0000 1101 0011 */
    { 12, 0xD4, 36 },   /* 0000 1101 0100 */
    { 12, 0xD5, 37 },   /* 0000 1101 0101 */
    { 12, 0xD6, 38 },   /* 0000 1101 0110 */
    { 12, 0xD7, 39 },   /* 0000 1101 0111 */
    { 12, 0x6C, 40 },   /* 0000 0110 1100 */
    { 12, 0x6D, 41 },   /* 0000 0110 1101 */
    { 12, 0xDA, 42 },   /* 0000 1101 1010 */
    { 12, 0xDB, 43 },   /* 0000 1101 1011 */
    { 12, 0x54, 44 },   /* 0000 0101 0100 */
    { 12, 0x55, 45 },   /* 0000 0101 0101 */
    { 12, 0x56, 46 },   /* 0000 0101 0110 */
    { 12, 0x57, 47 },   /* 0000 0101 0111 */
    { 12, 0x64, 48 },   /* 0000 0110 0100 */
    { 12, 0x65, 49 },   /* 0000 0110 0101 */
    { 12, 0x52, 50 },   /* 0000 0101 0010 */
    { 12, 0x53, 51 },   /* 0000 0101 0011 */
    { 12, 0x24, 52 },   /* 0000 0010 0100 */
    { 12, 0x37, 53 },   /* 0000 0011 0111 */
    { 12, 0x38, 54 },   /* 0000 0011 1000 */
    { 12, 0x27, 55 },   /* 0000 0010 0111 */
    { 12, 0x28, 56 },   /* 0000 0010 1000 */
    { 12, 0x58, 57 },   /* 0000 0101 1000 */
    { 12, 0x59, 58 },   /* 0000 0101 1001 */
    { 12, 0x2B, 59 },   /* 0000 0010 1011 */
    { 12, 0x2C, 60 },   /* 0000 0010 1100 */
    { 12, 0x5A, 61 },   /* 0000 0101 1010 */
    { 12, 0x66, 62 },   /* 0000 0110 0110 */
    { 12, 0x67, 63 },   /* 0000 0110 0111 */
    { 10, 0xF, 64 },    /* 0000 0011 11 */
    { 12, 0xC8, 128 },  /* 0000 1100 1000 */
    { 12, 0xC9, 192 },  /* 0000 1100 1001 */
    { 12, 0x5B, 256 },  /* 0000 0101 1011 */
    { 12, 0x33, 320 },  /* 0000 0011 0011 */
    { 12, 0x34, 384 },  /* 0000 0011 0100 */
    { 12, 0x35, 448 },  /* 0000 0011 0101 */
    { 13, 0x6C, 512 },  /* 0000 0011 0110 0 */
    { 13, 0x6D, 576 },  /* 0000 0011 0110 1 */
    { 13, 0x4A, 640 },  /* 0000 0010 0101 0 */
    { 13, 0x4B, 704 },  /* 0000 0010 0101 1 */
    { 13, 0x4C, 768 },  /* 0000 0010 0110 0 */
    { 13, 0x4D, 832 },  /* 0000 0010 0110 1 */
    { 13, 0x72, 896 },  /* 0000 0011 1001 0 */
    { 13, 0x73, 960 },  /* 0000 0011 1001 1 */
    { 13, 0x74, 1024 }, /* 0000 0011 1010 0 */
    { 13, 0x75, 1088 }, /* 0000 0011 1010 1 */
    { 13, 0x76, 1152 }, /* 0000 0011 1011 0 */
    { 13, 0x77, 1216 }, /* 0000 0011 1011 1 */
    { 13, 0x52, 1280 }, /* 0000 0010 1001 0 */
    { 13, 0x53, 1344 }, /* 0000 0010 1001 1 */
    { 13, 0x54, 1408 }, /* 0000 0010 1010 0 */
    { 13, 0x55, 1472 }, /* 0000 0010 1010 1 */
    { 13, 0x5A, 1536 }, /* 0000 0010 1101 0 */
    { 13, 0x5B, 1600 }, /* 0000 0010 1101 1 */
    { 13, 0x64, 1664 }, /* 0000 0011 0010 0 */
    { 13, 0x65, 1728 }, /* 0000 0011 0010 1 */
    { 11, 0x8, 1792 },  /* 0000 0001 000 */
    { 11, 0xC, 1856 },  /* 0000 0001 100 */
    { 11, 0xD, 1920 },  /* 0000 0001 101 */
    { 12, 0x12, 1984 }, /* 0000 0001 0010 */
    { 12, 0x13, 2048 }, /* 0000 0001 0011 */
    { 12, 0x14, 2112 }, /* 0000 0001 0100 */
    { 12, 0x15, 2176 }, /* 0000 0001 0101 */
    { 12, 0x16, 2240 }, /* 0000 0001 0110 */
    { 12, 0x17, 2304 }, /* 0000 0001 0111 */
    { 12, 0x1C, 2368 }, /* 0000 0001 1100 */
    { 12, 0x1D, 2432 }, /* 0000 0001 1101 */
    { 12, 0x1E, 2496 }, /* 0000 0001 1110 */
    { 12, 0x1F, 2560 }}; /* 0000 0001 1111 */

static sff_progress_handler _sff_encode_handler = NULL;
static sff_progress_handler _sff_decode_handler = NULL;

static WORD getbuf;
static WORD sffbytes,sffptr;
static BYTE sffbuffer[10000];

static WORD putbuf;
static BYTE putval;

static DWORD prev_offset;

static unsigned char sff_get_byte(HANDLE handle)
{
    DWORD read;
    
    if (sffbytes==0) {
        ReadFile(handle,sffbuffer,10000,&read,NULL);
        sffbytes=read;
        sffptr=0;
    } 
    sffbytes--;
    return(sffbuffer[sffptr++]);
}

// lsb bit first
static BYTE sff_get_bit(HANDLE handle)
{
    static BYTE getval;
    
    if (getbuf==0) {
        getval=sff_get_byte(handle);
        getbuf=8;
    }
    getbuf--;
    return(getval>>(7-getbuf)&0x1);
}    

static void sff_put_byte(HANDLE handle, BYTE b)
{
    DWORD written;
    
    if (sffptr==10000) {
        WriteFile(handle,sffbuffer,10000,&written,NULL);
        sffptr=0;
    } 
    sffbuffer[sffptr++]=b;
}

static void sff_flush_buffer(HANDLE handle)
{
    DWORD written;
    
    WriteFile(handle,sffbuffer,sffptr,&written,NULL);
    sffptr=0;
}

// lsb bit first
static void sff_put_bit(HANDLE handle, BYTE bit)
{    
    if (putbuf==0) sff_flush_bitbuffer(handle);
    putbuf--;
    putval|=( (bit&0x1)<<(7-putbuf) );
}    

static void sff_flush_bitbuffer(HANDLE handle)
{
    sff_put_byte(handle, putval);
    putval=0;
    putbuf=8;
}

static void sff_put_bits(HANDLE handle, int val, int number)
{
    if (number==0) return;
    do {
        number--;
        sff_put_bit(handle, val>>number&01);
    } while (number!=0);
}

void SFFencode_init(HANDLE handle)
{
    DWORD written;
    
    SFF_HEADER.SFF_Id=0x66666653;
    SFF_HEADER.Version=0x01;
    SFF_HEADER.reserved=0x00;
    SFF_HEADER.UserInformation=0x0000;
    SFF_HEADER.PageCount=0x0000;
    SFF_HEADER.OffsetFirstPageHeader=0x14;
    SFF_HEADER.OffsetLastPageHeader=0x00000000;
    SFF_HEADER.OffsetDocumentEnd=0x00000000;

    prev_offset=1;

    WriteFile(handle,&SFF_HEADER,sizeof(SFF_HEADER),&written,NULL);
}

void SFFencode_deinit(HANDLE handle)
{
    DWORD read,written,next_offset;
    int i;
    
    sffptr=0;
    sff_put_byte(handle, 254);
    sff_put_byte(handle, 0);
    sff_flush_buffer(handle);

    SetFilePointer(handle,0,NULL,FILE_BEGIN);
    WriteFile(handle,&SFF_HEADER,sizeof(SFF_HEADER),&written,NULL);

    prev_offset=SFF_HEADER.OffsetLastPageHeader;
    next_offset=0x1;
    for (i=0;i<SFF_HEADER.PageCount;i++) {
        SetFilePointer(handle,prev_offset,NULL,FILE_BEGIN);
        ReadFile(handle,&SFF_PAGE,sizeof(SFF_PAGE),&read,NULL);
        SFF_PAGE.OffsetNextPage=next_offset;
        SetFilePointer(handle,prev_offset,NULL,FILE_BEGIN);
        WriteFile(handle,&SFF_PAGE,sizeof(SFF_PAGE),&written,NULL);

        next_offset=prev_offset;
        prev_offset=SFF_PAGE.OffsetPreviousPage;
    }
}

void SFFencode(HANDLE handle, BYTE *image, WORD width, WORD height, WORD bpp)
{
    int x,y,o,i;
    DWORD written;
    int white, real_width=width;
    int white_skip=0, white_line;
    int codelist[2000]; // should never contain more than +/- 1728 codes...
    int codes, mhlength, length;
    BYTE mhcodelist[32768];
    int mhcodes;
    int pixel;

    if (bpp!=1) return;

    SFF_HEADER.PageCount++;
    SFF_HEADER.OffsetLastPageHeader=SetFilePointer(handle,0,NULL,FILE_CURRENT);
    
    SFF_PAGE.PageHeaderId=254;
    SFF_PAGE.PageHeaderLen=0x10;
    SFF_PAGE.ResolutionVertical=0;
    SFF_PAGE.ResolutionHorizontal=0;
    SFF_PAGE.Coding=0;
    SFF_PAGE.reserved=0;
    SFF_PAGE.LineLength=1728;
    SFF_PAGE.PageLength=height;
    SFF_PAGE.OffsetPreviousPage=prev_offset;
    SFF_PAGE.OffsetNextPage=0x00000000;

    prev_offset=SFF_HEADER.OffsetLastPageHeader;

    WriteFile(handle,&SFF_PAGE,sizeof(SFF_PAGE),&written,NULL);

    if (width>1728) width=1728;

    sffptr=0;
    for (y=0;y<height;y++) {
        if (height!=0) sff_encode_progress(100*y/height);

        white=1;
        codes=1;
        codelist[codes]=0;

        white_line=1;
        for (x=0;x<width;x++) {
            pixel=image[(y*real_width+x)*3]>128;
            if (pixel==0) white_line=0;

            if (pixel && white) codelist[codes]++; else
            if (!pixel && white) { white=0; codes++; codelist[codes]=1; } else
            if (pixel && !white) { white=1; codes++; codelist[codes]=1; } else
            if (!pixel && !white) codelist[codes]++;
        }

        if (real_width<1728) {
            if (!white) { codes++; codelist[codes]=1728-real_width; }
            if (white) codelist[codes]+=1728-real_width;
        }

        if (white_line) white_skip++;
        if (white_skip==37 ||                // max white skip
            (!white_line && white_skip) ||   // no more white lines
            (white_skip && y==height-1)) {   // last line in image
            sff_put_byte(handle, 216+white_skip);
            white_skip=0;
        }

        if (!white_line) {
            mhcodes=0;
            
            white=1;
            mhlength=0;
            for (o=1;o<=codes;o++) {
                do {
                    if (white) for (i=104-1;i>=0;i--) {
                        if (codelist[o]>=Group3FaxWhiteCodes[i].runlen) {
                            mhcodelist[mhcodes++]=i;
                            codelist[o]-=Group3FaxWhiteCodes[i].runlen;
                            mhlength+=Group3FaxWhiteCodes[i].length;

                            if (codelist[o]==0 && Group3FaxWhiteCodes[i].runlen>63) {
                                mhcodelist[mhcodes++]=0;
                                mhlength+=Group3FaxWhiteCodes[0].length;
                            }
                            break;
                        }
                    }

                    if (!white) for (i=104-1;i>=0;i--) {
                        if (codelist[o]>=Group3FaxBlackCodes[i].runlen) {
                            mhcodelist[mhcodes++]=i;
                            codelist[o]-=Group3FaxBlackCodes[i].runlen;
                            mhlength+=Group3FaxBlackCodes[i].length;

                            if (codelist[o]==0 && Group3FaxBlackCodes[i].runlen>63) {
                                mhcodelist[mhcodes++]=0;
                                mhlength+=Group3FaxBlackCodes[0].length;
                            }
                            break;
                        }
                    }
                } while (codelist[o]);
                white=!white;
            }

            length=mhlength/8;
            if (mhlength%8) length++;

            if (length<=216) {
                sff_put_byte(handle, length);
            } else {
                sff_put_byte(handle, 0);
                sff_put_byte(handle, length&0xff);
                sff_put_byte(handle, (length>>8)&0xff);
            }

            sff_flush_buffer(handle);

            putval=0;
            putbuf=8;

            white=1;
            for (o=0;o<mhcodes;o++) {
                if (white) {
                    sff_put_bits(handle,
                      Group3FaxWhiteCodes[mhcodelist[o]].code,
                      Group3FaxWhiteCodes[mhcodelist[o]].length);
                    if (Group3FaxWhiteCodes[mhcodelist[o]].runlen<64) white=0;
                } else {
                    sff_put_bits(handle,
                      Group3FaxBlackCodes[mhcodelist[o]].code,
                      Group3FaxBlackCodes[mhcodelist[o]].length);
                    if (Group3FaxBlackCodes[mhcodelist[o]].runlen<64) white=1;
                }
            }

            sff_flush_bitbuffer(handle);
        }

    }

    sff_flush_buffer(handle);
}

void SFFdecode(int page, HANDLE handle, BYTE **image)
{
    int x,y,i,o;
    BYTE *ptr=*image;
    WORD width, height, bpp;
    BYTE code;
    WORD lcode;

    if (SFFinfo(page, handle, &width, &height, &bpp)<0) return;
    sffbytes=0;

    if (SFF_PAGE.Coding==0) {    // Modified Huffman
        y=0;
        while ((code=sff_get_byte(handle))!=254) {
            x=0;
            if (code<=216) { // pixel row
                int stop, count, run, white, woord, bitsleft;

                if (y>height) return;

                if (code==0) {
                    BYTE b1, b2;
                    b1=sff_get_byte(handle);
                    b2=sff_get_byte(handle);
                    lcode=b2<<8|b1;
                } else lcode=code;

                getbuf=0;
                white=1;
                bitsleft=lcode*8;

                while (x<width && bitsleft>3) {
                    woord=sff_get_bit(handle);bitsleft--;
                    count=1;
                    stop=0;
                    
                    do {
                        woord=(woord<<1)|sff_get_bit(handle);bitsleft--;
                        count++;
                        if (white) {
                            for (i=0;i<104;i++) {
                                if (woord==Group3FaxWhiteCodes[i].code && count==Group3FaxWhiteCodes[i].length) {
                                    run=Group3FaxWhiteCodes[i].runlen;
                                    stop=1;
                                    break;
                                }
                            }
                        } else {
                            for (i=0;i<104;i++) {
                                if (woord==Group3FaxBlackCodes[i].code && count==Group3FaxBlackCodes[i].length) {
                                    run=Group3FaxBlackCodes[i].runlen;
                                    stop=1;
                                    break;
                                }
                            }
                        }
                    } while (!stop && count!=14);

                    for (i=0;i<run;i++) {
                        if (x>width) return;
                        ptr[(y*width+x)*3+0]=
                        ptr[(y*width+x)*3+1]=
                        ptr[(y*width+x)*3+2]=white?255:0;
                        x++;
                    }
                    
                    if (run<64) white=!white;
                }

                for (i=0;i<bitsleft;i++) sff_get_bit(handle);
                y++;
                sff_decode_progress(100*y/height);
            } else
            if (code<=253) { // skip 1..37 lines
                for (i=0;i<code-216;i++) {
                    if (y>height) return;
                    for (o=0;o<width;o++) {
                        ptr[(y*width+o)*3+0]=
                        ptr[(y*width+o)*3+1]=
                        ptr[(y*width+o)*3+2]=255;
                    }
                    y++;
                    sff_decode_progress(100*y/height);
                }
            } else
            if (code==255) { // reserved extensions
                code=sff_get_byte(handle);
                if (code==0) {
                    if (y>height) return;
                    if (y>0) {
                        for (i=0;i<width;i++) {
                            if (x>width) return;
                            ptr[(y*width+x)*3+0]=ptr[((y-1)*width+x)*3+0];
                            ptr[(y*width+x)*3+1]=ptr[((y-1)*width+x)*3+1];
                            ptr[(y*width+x)*3+2]=ptr[((y-1)*width+x)*3+2];
                            x++;
                        }
                    } else {
                        for (i=0;i<width;i++) {
                            if (x>width) return;
                            ptr[(y*width+x)*3+0]=0;
                            ptr[(y*width+x)*3+1]=0;
                            ptr[(y*width+x)*3+2]=0;
                            x++;
                        }
                    }
                    y++;
                    sff_decode_progress(100*y/height);
                } else {
                    for (i=0;i<code;i++)
                        sff_get_byte(handle);
                }
            }
        } // while
    }
}

int SFFmoreinfo(WORD page, HANDLE handle, WORD *width, WORD *height, WORD *bpp, WORD *dpi_x, WORD *dpi_y)
{
    DWORD read;
    
    SetFilePointer(handle,0,NULL,FILE_BEGIN);
    ReadFile(handle,&SFF_HEADER,sizeof(SFF_HEADER),&read,NULL);
    if (SFF_HEADER.SFF_Id!=0x66666653) return(-255);
    if (SFF_HEADER.Version!=0x01) return(-255);
    if (SFF_HEADER.PageCount==0) SFF_HEADER.PageCount=1;

    if (page==0) {
        return(SFF_HEADER.PageCount);
    } else {
        int curpage=1;
        
        if (page>SFF_HEADER.PageCount) return(-254);
        SetFilePointer(handle,SFF_HEADER.OffsetFirstPageHeader,NULL,FILE_BEGIN);
        while (curpage!=page) {
            ReadFile(handle,&SFF_PAGE,sizeof(SFF_PAGE),&read,NULL);
            if (2+SFF_PAGE.PageHeaderLen!=sizeof(SFF_PAGE))
                SetFilePointer(handle,SFF_PAGE.PageHeaderLen+2-sizeof(SFF_PAGE),NULL,FILE_CURRENT);
            if (SFF_PAGE.PageHeaderId!=254) return(-253);
            SetFilePointer(handle,SFF_PAGE.OffsetNextPage,NULL,FILE_BEGIN);
            curpage++;
        }
        ReadFile(handle,&SFF_PAGE,sizeof(SFF_PAGE),&read,NULL);
        if (2+SFF_PAGE.PageHeaderLen!=sizeof(SFF_PAGE))
            SetFilePointer(handle,SFF_PAGE.PageHeaderLen+2-sizeof(SFF_PAGE),NULL,FILE_CURRENT);
        if (SFF_PAGE.PageHeaderId!=254) return(-253);

        if (width!=NULL) *width = SFF_PAGE.LineLength;
        if (height!=NULL) *height = SFF_PAGE.PageLength;
        if (dpi_x!=NULL) {
            switch (SFF_PAGE.ResolutionHorizontal) {
                case 0: *dpi_x=203;break;
                default: *dpi_x=0;
            }
        }
        if (dpi_y!=NULL) {
            switch (SFF_PAGE.ResolutionVertical) {
                case 0: *dpi_y=98;break;
                case 1: *dpi_y=196;break;
                default: *dpi_y=0;
            }
        }
        if (bpp!=NULL) *bpp = 1;

        if (SFF_PAGE.Coding!=0)
            return(-252);
         else
            return(1);
    }
}

int SFFinfo(WORD page, HANDLE handle, WORD *width, WORD *height, WORD *bpp)
{
    return(SFFmoreinfo(page, handle, width, height, bpp, NULL, NULL));
}

void sff_set_encode_handler(sff_progress_handler handler)
{
    _sff_encode_handler=handler;
}

void sff_set_decode_handler(sff_progress_handler handler)
{
    _sff_decode_handler=handler;
}

static void sff_encode_progress(int percentage)
{
    if (_sff_encode_handler) _sff_encode_handler(percentage);
}

static void sff_decode_progress(int percentage)
{
    if (_sff_decode_handler) _sff_decode_handler(percentage);
}
